from collections.abc import Sequence
from functools import singledispatch
from typing import Dict, List, Optional, Type, Union

import numpy
import pandas
from pandas.core.dtypes.inference import is_sequence
from pydantic import create_model

from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.rest.schema import (
    get_array_model,
    get_dataframe_model,
    get_mapping_model,
    get_nested_sequence_model,
    get_series_model,
    get_single_value_model,
)
from tktl.core.t import RestSchemaTypes


@singledispatch
def to_pydantic(
    value, name: str = "SingleValue"
) -> Type[CustomDeserializingModelT]:  # noqa
    single_value_model = get_single_value_model(value)
    single_value_model.__name__ = name
    return single_value_model


@to_pydantic.register
def _(
    value: pandas.DataFrame, name: str = RestSchemaTypes.DATAFRAME.value, nullable=True
):
    """Create a Pydantic model for a batch from a Pandas DataFrame

    Parameters
    ----------
    value : pd.DataFrame
        Input Dataframe
    name : str, optional
        Name for the model, by default "DataFrame"
    nullable : bool, optional
        Indicates whether observations can be missing (None), by default True

    Returns
    -------
    ModelMetaclass
        Pydantic model of the dataframe
    """
    type_map = {}
    for col, values in value.to_dict().items():
        types = (type(v) for k, v in values.items() if v is not None)
        var_type = next(types, str)  # type of first item that is not None
        if nullable:
            var_type = Optional[var_type]
        type_map[col] = (var_type, None)
    index_type = type(next((t for t in value.index)))
    type_map["index"] = (index_type, None)

    DynamicBaseModel = create_model(name, **type_map)  # noqa
    DynamicBaseModel.__name__ = name
    return get_dataframe_model(base_model=DynamicBaseModel)


@to_pydantic.register
def _(value: pandas.Series, name: str = RestSchemaTypes.SERIES.value):
    """Create a Pydantic model for a batch from a Pandas Series

    Parameters
    ----------
    value : pd.Series
        Input Dataframe
    name : str, optional
        Name for the model, by default "DataFrame"
    nullable : bool, optional
        Indicates whether observations can be missing (None), by default True

    Returns
    -------
    ModelMetaclass
        Pydantic model of the series
    """
    nullable = True if value.isna().sum() > 0 else False
    type_name = "Outcome" if not value.name else value.name
    type_map = {}
    types = (type(v) for k, v in value.to_dict().items() if v is not None)
    var_type = next(types, str)  # type of first item that is not None
    if var_type == bool:
        var_type = Union[var_type, float]
    if nullable:
        var_type = Optional[var_type]
    type_map[type_name] = (List[var_type], None)
    DynamicBaseModel = create_model(name, **type_map)  # noqa
    DynamicBaseModel.__name__ = name
    return get_series_model(series=value, base_model=DynamicBaseModel)


@to_pydantic.register
def _(value: numpy.ndarray, name: str = RestSchemaTypes.ARRAY.value):
    """Create a Pydantic model for a batch from a Pandas Series

    Parameters
    ----------
    value : numpy.ndarray
    name : str, optional
        Name for the model, by default "DataFrame"

    Returns
    -------
    ModelMetaclass
        Pydantic model of the series
    """
    as_list = value.tolist()
    if len(value.shape) == 1:
        return List[type(as_list[0])]
    else:
        list_model = get_nested_sequence_model(as_list)
        array_model = get_array_model(list_model)
        array_model.__name__ = name
    return array_model


@to_pydantic.register
def _(value: Sequence, name: str = RestSchemaTypes.SEQUENCE.value):
    """Create a Pydantic model for a batch from a Pandas Series

    Parameters
    ----------
    value : Union[List, Dict]
        Input values, either Dict or List
    name : str, optional
        Name for the model, by default "DataFrame"

    Returns
    -------
    ModelMetaclass
        Pydantic model of the series
    """

    if isinstance(value, list) or isinstance(value, tuple):
        if len(value) > 0:
            first = value[0]
            if isinstance(first, list):
                sequence_model = get_nested_sequence_model(first)
                sequence_model.__name__ = name

            elif isinstance(first, dict):
                mapping_model = _dict_model(
                    value=first, name=RestSchemaTypes.DICT.value
                )
                sequence_model = get_mapping_model(base_model=mapping_model)
            # more specific than numpy's is_sequence implementation
            elif not is_sequence(first):
                sequence_model = get_single_value_model(first)
                sequence_model.__name__ = name
            else:
                raise
        else:
            raise ValueError("Input must be non-empty")
    else:
        raise
    return sequence_model


@to_pydantic.register
def _(value: dict, name: str = RestSchemaTypes.DICT.value):
    mapping_model = _dict_model(value=value, name=name)
    mapping_model.__name__ = name
    return mapping_model


def _dict_model(value, name):
    types = {k: (type(v), v) for k, v in value.items()}
    return create_model(name, **types)



import numpy
import pandas
import pandas.api.types as ptypes
import pytest
from pydantic import ValidationError, parse_obj_as

from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.rest.typed import to_pydantic


def test_from_single_value():
    schema = to_pydantic(342314)
    assert schema.__name__ == "SingleValue"
    assert schema.validate({"value": 1})
    deser = schema(**{"value": 1}).deserialize()
    assert deser == 1


def test_from_series(serializer_df_inputs: pandas.DataFrame):
    series = serializer_df_inputs.A
    schema = to_pydantic(series)
    assert schema.__name__ == "Series"
    rand_values = {"A": [1, 1, 12, 1, 1, 321, None]}
    assert schema.validate(rand_values)
    deser = schema(**rand_values).deserialize()
    assert isinstance(deser, pandas.Series)
    assert ptypes.is_float_dtype(deser.dtype)


def test_from_frame(serializer_df_inputs: pandas.DataFrame):
    schema = to_pydantic(serializer_df_inputs)
    assert schema.__name__ == "DataFrame"
    rand_values = [{k: numpy.random.randn(1)[0] for k in serializer_df_inputs.columns}]*10
    assert schema.validate(rand_values)
    deser = schema.parse_obj(rand_values).deserialize()
    assert isinstance(deser, pandas.DataFrame)
    for col in deser.columns:
        assert ptypes.is_float_dtype(deser[col].dtype)


def test_from_array_1d():
    arr = numpy.random.randn(100)
    schema = to_pydantic(arr)
    assert str(schema) == "typing.List[float]"
    assert isinstance(parse_obj_as(schema, [1]), list)
    assert isinstance(parse_obj_as(schema, [1, 10000.]), list)


def test_from_multi_dim_array():
    arr = numpy.random.randint(0, 100, size=(10, 5, 2, 2))
    schema = to_pydantic(arr)
    assert schema.__name__ == "Array"
    sample = numpy.random.randint(0, 100, size=(5, 2, 2)).tolist()
    assert schema.validate([sample])
    with pytest.raises(ValidationError):
        assert schema.validate([[[312, 432], [312, 432]]])

    deser = schema.parse_obj([sample]).deserialize()
    assert isinstance(deser, numpy.ndarray)


def test_from_sequence():
    sequence = numpy.random.randint(0, 5, size=(10, 2, 2)).tolist()
    schema = to_pydantic(sequence)
    with pytest.raises(ValidationError):
        schema.parse_obj([[1, 2], [3]])
    with pytest.raises(ValidationError):
        schema.parse_obj([[3, 1, 3]])

    parsed = schema.parse_obj([[1, 2], [3, 4]])
    assert isinstance(parsed, CustomDeserializingModelT)
    assert parsed.deserialize() == [[1, 2], [3, 4]]

    deeply_nested_sequence = numpy.random.randint(0, 5, size=(10, 2, 2, 2, 2)).tolist()
    schema = to_pydantic(deeply_nested_sequence)
    parsed = schema.parse_obj([
        [[[1, 2], [3, 4]], [[1, 2], [3, 4]]],
        [[[1, 2], [3, 4]], [[1, 2], [3, 4]]]
    ])
    assert parsed.deserialize() == [
        [[[1, 2], [3, 4]], [[1, 2], [3, 4]]],
        [[[1, 2], [3, 4]], [[1, 2], [3, 4]]]
    ]


def test_from_dict_or_list_of_dicts():
    sequence = [{"a": 1, "b": 2}, {"a": 1, "b": 100}]
    mapping_model = to_pydantic(sequence)
    valid = mapping_model(**{'a': 10, "b": 20})
    assert valid.deserialize() == {'a': 10, "b": 20}

    with pytest.raises(ValidationError):
        print(mapping_model(**{"c": 20, "b": "sdfgasdfg"}))



import logging

import click

from tktl.commands.validate import validate_project_config

_logger = logging.getLogger("root")


@click.command()
@click.option(
    "--path", "-p", help="Directory where the new project will be created", default="."
)
def validate(path) -> None:
    """Validates a new project for the necessary scaffolding, as well as the supporting
    files needed. The directory structure of a new project, and the files within it
    should look like this::

        regression.py         # Regression starter file
        classification.py     # Classification starter file
        tktl.yaml             # Main project config file
        .gitignore            # Just in case
        Dockerfile.tfserving  # TFS' Dockerfile]
        Dockerfile.racket     # racket's Dockerfile. To be implemented
        docker-compose.yaml   # Docker-compose to start up TFS

    """

    validate_project_config(path=path)



import json
from collections.abc import Sequence
from datetime import datetime
from typing import Any, List, Type, Union

import numpy
import pandas
from pydantic import BaseModel, conlist, validator

from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.utils import get_list_shape


def get_single_value_model(_value: Any) -> Type[CustomDeserializingModelT]:
    class SingleValueModel(CustomDeserializingModelT):
        value: type(_value)

        def deserialize(self):
            return self.value

    return SingleValueModel


def get_dataframe_model(base_model: Type[BaseModel]) -> Type[CustomDeserializingModelT]:
    validates_dt_model = _dt_validator_model(base_model)

    class DataFrame(CustomDeserializingModelT):
        __root__: List[validates_dt_model]

        def __iter__(self):
            return iter(self.__root__)

        def deserialize(self):
            values = self.dict()["__root__"]
            index = [v.pop("index") for v in values]
            if not all(index):
                return pandas.DataFrame.from_records(values)
            else:
                return pandas.DataFrame.from_records(
                    values, index=pandas.Index(data=index)
                )

    return DataFrame


def get_series_model(
    series: pandas.Series, base_model: Type[BaseModel]
) -> Type[CustomDeserializingModelT]:
    class Series(base_model, CustomDeserializingModelT):
        _name = series.name
        _dtype = series.dtype

        def deserialize(self):
            values = self.dict()[self._name]
            _series = pandas.Series(values)
            _series = _series.astype(self._dtype)  # column types
            return _series

    return Series


def get_jsonable_encoder_sequence_model(
    base_model: Type[BaseModel],
) -> Type[CustomDeserializingModelT]:
    class JsonableEncoderModel(base_model, CustomDeserializingModelT):
        def deserialize(self):
            return json.loads(self.values)

    return JsonableEncoderModel


def get_nested_sequence_model(sequence: Sequence) -> Type[CustomDeserializingModelT]:
    shape, types = get_list_shape(sequence)
    shape = (None,) + shape[1:]
    inner_model = conlist(Union[types], max_items=shape[-1], min_items=shape[-1])
    for s in reversed(shape[:-1]):
        inner_model = conlist(inner_model, max_items=s, min_items=s)
    return _nested_sequence_model(inner_models=inner_model)


def get_array_model(base_model) -> Type[CustomDeserializingModelT]:
    class ArrayModel(base_model, CustomDeserializingModelT):
        def deserialize(self):
            return numpy.array(self.dict()["__root__"])

    return ArrayModel


def get_mapping_model(base_model: Type[BaseModel]) -> Type[CustomDeserializingModelT]:
    class MappingModel(base_model, CustomDeserializingModelT):
        def deserialize(self):
            return self.dict()

    return MappingModel


def _dt_validator_model(base_model):
    class DtValidateModel(base_model):
        @validator("*", each_item=True, allow_reuse=True)
        def remove_tz(cls, v):
            if isinstance(v, datetime):
                return v.replace(tzinfo=None)
            return v

    return DtValidateModel


class SequenceEncoderModel(CustomDeserializingModelT):
    __root__: List

    def __iter__(self):
        return iter(self.__root__)

    def deserialize(self):
        return self.dict()["__root__"]


def get_typed_sequence_model(inner_type):
    class SimpleTypedSequenceEncoderModel(CustomDeserializingModelT):
        __root__: List[inner_type]

        def __iter__(self):
            return iter(self.__root__)

        def deserialize(self):
            return self.dict()["__root__"]

    return SimpleTypedSequenceEncoderModel


def _nested_sequence_model(inner_models: Type[BaseModel]):
    class NestedSequenceModel(SequenceEncoderModel):
        __root__: inner_models

    return NestedSequenceModel



from pyarrow.flight import (
    ClientAuthHandler,
    FlightClient,
    FlightDescriptor,
    FlightInfo,
    Ticket,
)

from tktl.core.clients import Client
from tktl.core.serializers import deserialize_arrow, serialize_arrow


class ApiKeyClientAuthHandler(ClientAuthHandler):
    """An example implementation of authentication via ApiKey."""

    def __init__(self, api_key: str):
        super(ApiKeyClientAuthHandler, self).__init__()
        self.api_key = api_key

    def authenticate(self, outgoing, incoming):
        outgoing.write(self.api_key)
        self.api_key = incoming.read()

    def get_token(self):
        return self.api_key


class ArrowFlightClient(Client):
    def list_endpoints(self):
        pass

    def list_deployments(self):
        pass

    def get_sample_data(self):
        pass

    framework = "Arrow Flight RPC"
    scheme = "grpc+tcp"

    def __init__(self, host, port, api_key: str):
        super().__init__(api_key)
        self.location = f"{self.scheme}://{host}:{port}"
        self.client = FlightClient(self.location)

    def list_commands(self):
        return self.client.list_actions()

    def predict(self, inputs, endpoint_name: str):
        table = serialize_arrow(inputs)
        descriptor = self.get_flight_info(command_name=str.encode(endpoint_name))
        writer, reader = self.client.do_exchange(descriptor)
        with writer:
            writer.begin(table.schema)
            writer.write_table(table)
            writer.done_writing()
            table = reader.read_all()
        return deserialize_arrow(table)

    def sample_data(self, command_name: str):
        x_info = self.client.do_get(Ticket(ticket=str.encode(f"{command_name}__X")))
        y_info = self.client.do_get(Ticket(ticket=str.encode(f"{command_name}__y")))
        return (
            deserialize_arrow(x_info.read_all()),
            deserialize_arrow(y_info.read_all()),
        )

    def get_schema(self, command_name: str):
        info = self.get_flight_info(str.encode(command_name))
        return info.schema

    def get_flight_info(self, command_name: bytes) -> FlightInfo:
        descriptor = FlightDescriptor.for_command(command_name)
        return self.client.get_flight_info(descriptor)



import pandas
from numpy.testing import assert_array_equal

from pydantic import BaseModel

from tktl.core.serializers import to_pydantic
from tktl.core.serializers.rest import DataFrameSerializer, SeriesSerializer


def test_frame_serializer(serializer_df_inputs: pandas.DataFrame):
    model = to_pydantic(serializer_df_inputs)
    as_pydantic_model = DataFrameSerializer.serialize(serializer_df_inputs, output_model=model)
    assert isinstance(as_pydantic_model, list)
    assert as_pydantic_model[0].index == serializer_df_inputs.index.values[0]
    serialized = model.parse_obj(serializer_df_inputs.to_dict('recods'))
    deser = serialized.deserialize()
    assert isinstance(deser, pandas.DataFrame)
    assert list(deser.columns) == list(serializer_df_inputs.columns)
    assert_array_equal(deser['A'].values, serializer_df_inputs['A'].values)
    assert set(serializer_df_inputs.columns.tolist()).issubset(set(serialized.dict()['__root__'][0].keys()))


def test_series_serializer(serializer_series_inputs: pandas.Series):
    model = to_pydantic(serializer_series_inputs)
    as_pydantic_model = SeriesSerializer.serialize(serializer_series_inputs, output_model=model)
    assert isinstance(as_pydantic_model, BaseModel)
    assert isinstance(as_pydantic_model.dict()["B"], list)



from typing import List

from .. import loggers as sdk_logger
from ..config import settings
from ..exceptions import TaktileSdkError
from ..managers.auth import AuthConfigManager
from ..schemas.repository import Endpoint, Repository, RepositoryDeployment
from ..utils import flatten, lru_cache
from .http_client import API, interpret_response


class TaktileClient(API):

    SCHEME: str

    def __init__(self, api_url, logger=sdk_logger.MuteLogger()):
        """
        Base class. All client classes inherit from it.
        """
        super().__init__(api_url, logger=logger)
        self.api_url = api_url
        self.api_key = AuthConfigManager.get_api_key()
        self.logger = logger

    @lru_cache(timeout=50, typed=False)
    def _get_repositories(self) -> List[Repository]:
        response = self.get(f"{settings.API_V1_STR}/models")
        return interpret_response(response=response, model=Repository)

    def get_repositories(self):
        repositories = self._get_repositories()
        return repositories

    def get_deployments_for_repository(
        self, repository_name: str, repository_owner: str = None
    ) -> List[RepositoryDeployment]:
        repos = [
            r for r in self.get_repositories() if r.repository_name == repository_name
        ]
        if not repos:
            raise TaktileSdkError(f"No repositories with name {repository_name} found")
        if len(repos) > 1:
            if not repository_owner:
                raise TaktileSdkError(
                    f"More than one repository with name {repository_name} found, please specify repository_owner"
                )
            try:
                return [r for r in repos if r.repository_owner == repository_owner][
                    0
                ].deployments
            except IndexError:
                raise TaktileSdkError(
                    f"No repository named {repository_name} with owner {repository_owner}"
                )
        else:
            return repos[0].deployments

    def get_endpoints_for_deployment(
        self, repository_name: str, repository_owner: str = None
    ) -> List[Endpoint]:
        deployments = self.get_deployments_for_repository(
            repository_name=repository_name, repository_owner=repository_owner
        )
        return list(flatten([r.endpoints for r in deployments]))

    def get_endpoint(
        self, endpoint_name: str, repository_name: str, repository_owner: str = None
    ):
        endpoints = [
            e
            for e in self.get_endpoints_for_deployment(
                repository_name=repository_name, repository_owner=repository_owner
            )
            if e.name == endpoint_name
        ]
        if not endpoints:
            raise TaktileSdkError(
                f"No endpoints named {endpoint_name} found for repository named {repository_name} "
                f"with owner {repository_owner}"
            )



from tktl.core.clients import Client


class RestClient(Client):

    TRANSPORT = "rest"

    def __init__(self, api_key: str):
        super().__init__(api_key)

    def predict(self):
        pass

    def list_endpoints(self):
        pass

    def list_deployments(self):
        pass

    def get_sample_data(self):
        pass

    def get_schema(self):
        pass



from distutils.version import StrictVersion

from tktl import __version__
from tktl.version import TaktileVersionChecker, VersionChecker


def test_get_version(capsys):
    TaktileVersionChecker.look_for_new_version_with_timeout()
    out, err = capsys.readouterr()
    assert out == ""
    vc = VersionChecker()
    up_to_date, version_from_repository = vc.is_up_to_date("taktile-cli", __version__)
    assert up_to_date is True
    assert StrictVersion(__version__) >= StrictVersion(version_from_repository)
